# Unit tests for event_manager.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements
# pylint: disable=too-few-public-methods
# pylint: disable=protected-access
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-locals
# pylint: disable=too-many-lines

"""Unit tests for event_manager.py methods."""

from __future__ import annotations

import itertools
import queue
from typing import TYPE_CHECKING
from unittest.mock import call

import gi
import pytest

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext
    from unittest.mock import MagicMock

@pytest.mark.unit
class TestEventManager:
    """Test EventManager class methods."""

    def _setup_ignore_event_ax_utilities_mocks(self, test_context, mock_time=5000.0, is_text=True):
        """Set up common AXUtilities mocks for event ignore testing scenarios."""

        test_context.patch("time.time", return_value=mock_time)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text", return_value=is_text
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_section", return_value=False
        )

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, MagicMock]:
        """Returns dependencies for event_manager module testing."""

        additional_modules = [
            "orca.input_event_manager",
            "orca.ax_utilities_debugging",
            "orca.ax_utilities",
        ]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)

        debug_mock = essential_modules["orca.debug"]
        debug_mock.LEVEL_INFO = 800
        debug_mock.LEVEL_WARNING = 2
        debug_mock.LEVEL_SEVERE = 3
        debug_mock.debugLevel = 0

        braille_mock = essential_modules["orca.braille"]
        braille_mock.disableBraille = test_context.Mock()

        focus_manager_mock = essential_modules["orca.focus_manager"]
        focus_mgr_instance = test_context.Mock()
        focus_mgr_instance.get_locus_of_focus = test_context.Mock()
        focus_mgr_instance.get_active_window = test_context.Mock()
        focus_mgr_instance.focus_and_window_are_unknown = test_context.Mock(
            return_value=False
        )
        focus_mgr_instance.clear_state = test_context.Mock()
        focus_manager_mock.get_manager = test_context.Mock(return_value=focus_mgr_instance)

        input_event_mock = essential_modules["orca.input_event"]

        class MockKeyboardEvent:
            """Mock KeyboardEvent for testing."""

        class MockBrailleEvent:
            """Mock BrailleEvent for testing."""

        class MockMouseButtonEvent:
            """Mock MouseButtonEvent for testing."""

        input_event_mock.KeyboardEvent = MockKeyboardEvent
        input_event_mock.BrailleEvent = MockBrailleEvent
        input_event_mock.MouseButtonEvent = MockMouseButtonEvent

        input_event_manager_mock = essential_modules["orca.input_event_manager"]
        input_mgr_instance = test_context.Mock()
        input_mgr_instance.start_key_watcher = test_context.Mock()
        input_mgr_instance.stop_key_watcher = test_context.Mock()
        input_event_manager_mock.get_manager = test_context.Mock(
            return_value=input_mgr_instance
        )

        script_manager_mock = essential_modules["orca.script_manager"]
        script_mgr_instance = test_context.Mock()
        script_instance = test_context.Mock()
        script_instance.app = test_context.Mock()
        script_instance.event_cache = {}
        script_instance.listeners = {}
        script_instance.is_activatable_event = test_context.Mock(return_value=True)
        script_instance.force_script_activation = test_context.Mock(return_value=False)
        script_instance.present_if_inactive = False
        script_mgr_instance.get_active_script = test_context.Mock(
            return_value=script_instance
        )
        script_mgr_instance.get_script = test_context.Mock(return_value=script_instance)
        script_mgr_instance.set_active_script = test_context.Mock()
        script_mgr_instance.get_default_script = test_context.Mock(
            return_value=script_instance
        )
        script_mgr_instance.reclaim_scripts = test_context.Mock()
        script_manager_mock.get_manager = test_context.Mock(return_value=script_mgr_instance)

        settings_mock = essential_modules["orca.settings"]
        settings_mock.progressBarVerbosity = 1
        settings_mock.PROGRESS_BAR_ALL = 2

        ax_utils_debugging_mock = essential_modules["orca.ax_utilities_debugging"]
        ax_utils_debugging_mock.object_event_details_as_string = test_context.Mock(
            return_value="mock details"
        )

        ax_object_mock = essential_modules["orca.ax_object"]
        ax_object_mock.get_name = test_context.Mock()
        ax_object_mock.get_parent = test_context.Mock()
        ax_object_mock.get_role = test_context.Mock()
        ax_object_mock.get_attribute = test_context.Mock()
        ax_object_mock.is_dead = test_context.Mock(return_value=False)
        ax_object_mock.has_state = test_context.Mock(return_value=False)

        ax_utilities_mock = essential_modules["orca.ax_utilities"]
        ax_utilities_mock.is_frame = test_context.Mock(return_value=False)
        ax_utilities_mock.is_dialog_or_alert = test_context.Mock(return_value=False)
        ax_utilities_mock.is_window = test_context.Mock(return_value=False)
        ax_utilities_mock.is_text = test_context.Mock(return_value=False)
        ax_utilities_mock.is_notification = test_context.Mock(return_value=False)
        ax_utilities_mock.is_alert = test_context.Mock(return_value=False)
        ax_utilities_mock.is_selected = test_context.Mock(return_value=False)
        ax_utilities_mock.is_focused = test_context.Mock(return_value=False)
        ax_utilities_mock.manages_descendants = test_context.Mock(return_value=False)
        ax_utilities_mock.is_section = test_context.Mock(return_value=False)
        ax_utilities_mock.get_application = test_context.Mock()
        ax_utilities_mock.get_desktop = test_context.Mock()
        ax_utilities_mock.is_invalid_role = test_context.Mock(return_value=False)
        ax_utilities_mock.is_menu_related = test_context.Mock(return_value=False)
        ax_utilities_mock.is_image = test_context.Mock(return_value=False)
        ax_utilities_mock.is_showing = test_context.Mock(return_value=True)
        ax_utilities_mock.is_selectable = test_context.Mock(return_value=False)
        ax_utilities_mock.is_menu = test_context.Mock(return_value=False)
        ax_utilities_mock.is_focusable = test_context.Mock(return_value=False)
        ax_utilities_mock.is_panel = test_context.Mock(return_value=False)
        ax_utilities_mock.is_modal = test_context.Mock(return_value=False)
        ax_utilities_mock.is_progress_bar = test_context.Mock(return_value=False)
        ax_utilities_mock.is_defunct = test_context.Mock(return_value=False)
        ax_utilities_mock.is_application_in_desktop = test_context.Mock(return_value=True)
        ax_utilities_mock.is_iconified = test_context.Mock(return_value=False)

        glib_mock = test_context.Mock()
        glib_mock.idle_add = test_context.Mock(return_value=123)
        glib_mock.timeout_add = test_context.Mock()
        test_context.patch("gi.repository.GLib", new=glib_mock)

        essential_modules["focus_manager_instance"] = focus_mgr_instance
        essential_modules["input_event_manager_instance"] = input_mgr_instance
        essential_modules["script_manager_instance"] = script_mgr_instance
        essential_modules["script_instance"] = script_instance
        essential_modules["glib"] = glib_mock

        return essential_modules

    def test_init(self, test_context: OrcaTestContext) -> None:
        """Test EventManager.__init__."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        assert not manager._script_listener_counts
        assert manager._active is False
        assert manager._paused is False
        assert isinstance(manager._counter, itertools.count)
        assert isinstance(manager._event_queue, queue.PriorityQueue)
        assert manager._gidle_id == 0
        assert not manager._event_history

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "activate_manager",
                "operation": "activate",
                "initial_active": False,
                "initial_counts": {},
                "expected_active": True,
                "watcher_method": "start_key_watcher",
            },
            {
                "id": "deactivate_manager",
                "operation": "deactivate",
                "initial_active": True,
                "initial_counts": {"test": 1},
                "expected_active": False,
                "watcher_method": "stop_key_watcher",
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_activation_operations(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test EventManager activate/deactivate operations."""
        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        mock_get_input_mgr = test_context.Mock()
        test_context.patch(
            "orca.event_manager.input_event_manager.get_manager", new=mock_get_input_mgr
        )
        mock_input_mgr = test_context.Mock()
        mock_get_input_mgr.return_value = mock_input_mgr
        manager = EventManager()
        manager._active = case["initial_active"]
        manager._script_listener_counts = case["initial_counts"].copy()

        getattr(manager, case["operation"])()
        assert manager._active is case["expected_active"]

        if case["operation"] == "deactivate":
            assert not manager._script_listener_counts

        getattr(mock_input_mgr, case["watcher_method"]).assert_called_once()

        getattr(mock_input_mgr, case["watcher_method"]).reset_mock()
        getattr(manager, case["operation"])()
        getattr(mock_input_mgr, case["watcher_method"]).assert_not_called()

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "pause_only", "pause": True, "clear_queue": False, "reason": ""},
            {
                "id": "unpause_with_reason",
                "pause": False,
                "clear_queue": False,
                "reason": "test reason",
            },
            {
                "id": "pause_and_clear",
                "pause": True,
                "clear_queue": True,
                "reason": "clear and pause",
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_pause_queuing(self, test_context, case: dict) -> None:
        """Test EventManager.pause_queuing."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        original_queue = manager._event_queue
        manager.pause_queuing(case["pause"], case["clear_queue"], case["reason"])
        assert manager._paused == case["pause"]
        if case["clear_queue"]:
            assert manager._event_queue is not original_queue
        else:
            assert manager._event_queue is original_queue

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "window_event", "event_type": "window:activate", "expected_priority": 2},
            {
                "id": "focus_changed",
                "event_type": "object:state-changed:focused",
                "expected_priority": 3,
            },
            {
                "id": "active_descendant",
                "event_type": "object:active-descendant-changed",
                "expected_priority": 3,
            },
            {
                "id": "announcement_normal",
                "event_type": "object:announcement",
                "expected_priority": 4,
            },
            {
                "id": "invalid_entry",
                "event_type": "object:state-changed:invalid-entry",
                "expected_priority": 5,
            },
            {
                "id": "children_changed",
                "event_type": "object:children-changed:add",
                "expected_priority": 6,
            },
            {
                "id": "other_event",
                "event_type": "object:text-changed:insert",
                "expected_priority": 4,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_get_priority(self, test_context, case: dict) -> None:
        """Test EventManager._get_priority."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = case["event_type"]
        mock_event.source = test_context.Mock()
        mock_event.detail1 = 0
        ax_utilities = essential_modules["orca.ax_utilities"]
        if case["event_type"] == "object:state-changed:active":
            ax_utilities.is_frame.return_value = True
        elif case["event_type"] == "object:announcement":
            mock_event.detail1 = Atspi.Live.POLITE if case["expected_priority"] == 3 else 0
        priority = manager._get_priority(mock_event)
        assert priority == case["expected_priority"]

    def test_get_priority_announcement_levels(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._get_priority for announcement event levels."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:announcement"
        mock_event.source = test_context.Mock()

        mock_event.detail1 = Atspi.Live.ASSERTIVE
        priority = manager._get_priority(mock_event)
        assert priority == 2  # PRIORITY_IMPORTANT

        mock_event.detail1 = Atspi.Live.POLITE
        priority = manager._get_priority(mock_event)
        assert priority == 3  # PRIORITY_HIGH

        mock_event.detail1 = 999
        priority = manager._get_priority(mock_event)
        assert priority == 4  # PRIORITY_NORMAL

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "duplicate_events",
                "new_type": "object:text-changed:insert",
                "new_detail1": 5,
                "new_detail2": 10,
                "new_data": "test",
                "existing_type": "object:text-changed:insert",
                "existing_detail1": 5,
                "existing_detail2": 10,
                "existing_data": "test",
                "priority": 4,
                "should_obsolete": True,
            },
            {
                "id": "window_events",
                "new_type": "window:activate",
                "new_detail1": 0,
                "new_detail2": 0,
                "new_data": None,
                "existing_type": "window:deactivate",
                "existing_detail1": 0,
                "existing_detail2": 0,
                "existing_data": None,
                "priority": 2,
                "should_obsolete": True,
            },
            {
                "id": "no_obsolescence",
                "new_type": "object:text-changed:insert",
                "new_detail1": 0,
                "new_detail2": 0,
                "new_data": None,
                "existing_type": None,
                "existing_detail1": None,
                "existing_detail2": None,
                "existing_data": None,
                "priority": None,
                "should_obsolete": False,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_is_obsoleted_by_scenarios(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test EventManager._is_obsoleted_by for various obsolescence scenarios."""
        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = case["new_type"]
        mock_event.source = test_context.Mock()
        mock_event.detail1 = case["new_detail1"]
        mock_event.detail2 = case["new_detail2"]
        mock_event.any_data = case["new_data"]

        if case["existing_type"] is not None and case["priority"] is not None:
            existing_event = test_context.Mock(spec=Atspi.Event)
            existing_event.type = case["existing_type"]
            existing_event.source = mock_event.source
            existing_event.detail1 = case["existing_detail1"]
            existing_event.detail2 = case["existing_detail2"]
            existing_event.any_data = case["existing_data"]
            manager._event_queue.put((case["priority"], 1, existing_event))

        result = manager._is_obsoleted_by(mock_event)
        if case["should_obsolete"]:
            assert result is not None
        else:
            assert result is None

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "inactive_manager", "active": False, "paused": False, "expected": True},
            {"id": "paused_manager", "active": True, "paused": True, "expected": True},
        ],
        ids=lambda case: case["id"],
    )
    def test_ignore_inactive_or_paused_manager(
        self, test_context: OrcaTestContext, case: dict
    ) -> None:
        """Test EventManager._ignore when manager is inactive or paused."""
        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:text-changed:insert"

        manager._active = case["active"]
        manager._paused = case["paused"]
        assert manager._ignore(mock_event) is case["expected"]

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "window_activate_event",
                "event_type": "window:activate",
                "event_config": {},
                "expected_result": False,
            },
            {
                "id": "mouse_button_event",
                "event_type": "mouse:button:1p",
                "event_config": {},
                "expected_result": False,
            },
            {
                "id": "focused_window_event",
                "event_type": "object:state-changed:focused",
                "event_config": {"is_window": True},
                "expected_result": True,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_ignore_window_and_mouse_events(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test EventManager._ignore for window, mouse, and focused window events."""
        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = case["event_type"]
        mock_event.source = test_context.Mock()

        if "is_window" in case["event_config"]:
            ax_utilities = essential_modules["orca.ax_utilities"]
            ax_utilities.is_window.return_value = case["event_config"]["is_window"]

        assert manager._ignore(mock_event) is case["expected_result"]

    def test_ignore_frame_events(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore for frame events."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:test-event"
        mock_event.source = test_context.Mock()
        mock_event.any_data = test_context.Mock()
        mock_event.detail1 = 0
        mock_event.detail2 = 0
        ax_utilities = essential_modules["orca.ax_utilities"]
        focus_mgr = essential_modules["focus_manager_instance"]

        focus_mgr.get_locus_of_focus.return_value = None
        ax_utilities.is_window.return_value = False
        test_context.patch("time.time", return_value=1000.0)
        manager._event_history = {}

        mock_app = test_context.Mock()
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=True
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=mock_app
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_name", return_value="mutter-x11-frames"
        )
        assert manager._ignore(mock_event) is True

        regular_app = test_context.Mock()
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=True
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=regular_app
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_name", return_value="regular-app"
        )
        assert manager._ignore(mock_event) is False

    def test_ignore_text_events(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore for text-related events."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.source = test_context.Mock()
        mock_event.detail2 = 5001  # Large text insertion
        mock_event.any_data = test_context.Mock()
        mock_event.detail1 = 0
        ax_utilities = essential_modules["orca.ax_utilities"]
        focus_mgr = essential_modules["focus_manager_instance"]
        focus_mgr.get_locus_of_focus.return_value = None
        ax_utilities.is_window.return_value = False
        ax_utilities.is_frame.return_value = False
        ax_utilities.is_text.return_value = True  # This triggers text logic
        ax_utilities.is_notification.return_value = False
        ax_utilities.is_alert.return_value = False
        ax_utilities.is_selected.return_value = False
        ax_utilities.is_focused.return_value = False
        ax_utilities.is_section.return_value = False

        self._setup_ignore_event_ax_utilities_mocks(test_context, mock_time=5000.0, is_text=True)
        manager._event_history = {}

        mock_event.type = "object:text-changed:insert"
        result = manager._ignore(mock_event)
        assert result is True, f"Expected True but got {result}"

        mock_event.detail2 = 100
        assert manager._ignore(mock_event) is False

        mock_event.type = "object:text-changed:delete"
        assert manager._ignore(mock_event) is False

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "notification_event",
                "utility_method": "is_notification",
                "utility_value": True,
                "expected_result": False,
            },
            {
                "id": "alert_event",
                "utility_method": "is_alert",
                "utility_value": True,
                "expected_result": False,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_ignore_notification_and_alert_events(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test EventManager._ignore for notification and alert events."""
        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:test-event"
        mock_event.source = test_context.Mock()
        ax_utilities = essential_modules["orca.ax_utilities"]

        ax_utilities.is_notification.return_value = False
        ax_utilities.is_alert.return_value = False
        getattr(ax_utilities, case["utility_method"]).return_value = case["utility_value"]

        assert manager._ignore(mock_event) is case["expected_result"]

    def test_ignore_focus_and_selection_events(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore for focus and selection events."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:test-event"
        mock_event.source = test_context.Mock()
        mock_event.any_data = test_context.Mock()
        ax_utilities = essential_modules["orca.ax_utilities"]
        focus_mgr = essential_modules["focus_manager_instance"]

        focus_mgr.get_locus_of_focus.return_value = mock_event.source
        assert manager._ignore(mock_event) is False

        focus_mgr.get_locus_of_focus.return_value = mock_event.any_data
        assert manager._ignore(mock_event) is False

        focus_mgr.get_locus_of_focus.return_value = test_context.Mock()
        ax_utilities.is_selected.return_value = True
        assert manager._ignore(mock_event) is False

    def test_ignore_focused_source_events(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore for events from focused sources."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.source = test_context.Mock()
        mock_event.detail1 = 1
        mock_event.detail2 = 0
        mock_event.any_data = test_context.Mock()
        focus_mgr = essential_modules["focus_manager_instance"]

        focus_mgr.get_locus_of_focus.return_value = None
        test_context.patch("time.time", return_value=3000.0)
        manager._event_history = {}

        mock_event.type = "object:test-event"
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused", return_value=True
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.manages_descendants", return_value=False
        )
        assert manager._ignore(mock_event) is False

        # - should not be ignored
        mock_event.type = "object:state-changed:focused"
        mock_event.detail1 = 1
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused", return_value=True
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.manages_descendants", return_value=True
        )
        assert manager._ignore(mock_event) is False

    def test_ignore_live_region_events(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore for live region events."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:text-changed:insert"
        mock_event.source = test_context.Mock()
        mock_event.detail2 = 100
        ax_utilities = essential_modules["orca.ax_utilities"]
        ax_object = essential_modules["orca.ax_object"]
        focus_mgr = essential_modules["focus_manager_instance"]

        focus_mgr.get_locus_of_focus.return_value = test_context.Mock()
        ax_utilities.is_selected.return_value = False
        ax_utilities.is_focused.return_value = False
        ax_utilities.is_section.return_value = True

        ax_object.get_attribute.return_value = "polite"
        assert manager._ignore(mock_event) is False

        ax_object.get_attribute.return_value = "off"
        mock_app = test_context.Mock()
        ax_utilities.get_application.return_value = mock_app
        manager._event_history = {}
        assert manager._ignore(mock_event) is False

    def test_ignore_spam_filtering(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore spam filtering mechanism."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:test-event"
        mock_event.source = test_context.Mock()
        mock_event.detail1 = 0
        mock_event.detail2 = 0
        mock_event.any_data = test_context.Mock()
        mock_app = test_context.Mock()
        original_hash = hash

        def mock_hash(obj):
            if obj is mock_app:
                return 12345
            return original_hash(obj)

        test_context.patch("builtins.hash", new=mock_hash)
        focus_mgr = essential_modules["focus_manager_instance"]
        focus_mgr.get_locus_of_focus.return_value = None

        manager._event_history = {}
        test_context.patch("time.time", return_value=100.0)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_section", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.manages_descendants", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=mock_app
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_name", return_value="regular-app"
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_attribute", return_value=None
        )
        test_context.patch(
            "orca.event_manager.focus_manager.get_manager", return_value=focus_mgr
        )
        assert manager._ignore(mock_event) is False

        test_context.patch("time.time", return_value=100.05)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_section", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.manages_descendants", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=mock_app
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_name", return_value="regular-app"
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_attribute", return_value=None
        )
        test_context.patch(
            "orca.event_manager.focus_manager.get_manager", return_value=focus_mgr
        )
        assert manager._ignore(mock_event) is True

        test_context.patch("time.time", return_value=100.2)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_section", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.manages_descendants", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=mock_app
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_name", return_value="regular-app"
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_attribute", return_value=None
        )
        test_context.patch(
            "orca.event_manager.focus_manager.get_manager", return_value=focus_mgr
        )
        assert manager._ignore(mock_event) is False

    def test_ignore_mutter_events(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore for mutter-x11-frames events."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:test-event"
        mock_event.source = test_context.Mock()
        mock_event.detail1 = 0
        mock_event.detail2 = 0
        mock_event.any_data = test_context.Mock()
        mock_app = test_context.Mock()
        original_hash = hash

        def mock_hash(obj):
            if obj is mock_app:
                return 12345
            return original_hash(obj)

        test_context.patch("builtins.hash", new=mock_hash)
        focus_mgr = essential_modules["focus_manager_instance"]
        focus_mgr.get_locus_of_focus.return_value = None

        test_context.patch("time.time", return_value=100.0)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_section", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.manages_descendants", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=mock_app
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_name", return_value="mutter-x11-frames"
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_attribute", return_value=None
        )
        test_context.patch(
            "orca.event_manager.focus_manager.get_manager", return_value=focus_mgr
        )
        manager._event_history = {}
        assert manager._ignore(mock_event) is True

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "window_event",
                "event_type": "window:activate",
                "is_frame": False,
                "is_dialog_or_alert": False,
                "detail1": 0,
                "expected_priority": 2,
            },
            {
                "id": "active_frame",
                "event_type": "object:state-changed:active",
                "is_frame": True,
                "is_dialog_or_alert": False,
                "detail1": 0,
                "expected_priority": 2,
            },
            {
                "id": "active_dialog",
                "event_type": "object:state-changed:active",
                "is_frame": False,
                "is_dialog_or_alert": True,
                "detail1": 0,
                "expected_priority": 2,
            },
            {
                "id": "focused_event",
                "event_type": "object:state-changed:focused",
                "is_frame": False,
                "is_dialog_or_alert": False,
                "detail1": 0,
                "expected_priority": 3,
            },
            {
                "id": "active_descendant",
                "event_type": "object:active-descendant-changed",
                "is_frame": False,
                "is_dialog_or_alert": False,
                "detail1": 0,
                "expected_priority": 3,
            },
            {
                "id": "assertive_announcement",
                "event_type": "object:announcement",
                "is_frame": False,
                "is_dialog_or_alert": False,
                "detail1": 2,
                "expected_priority": 2,
            },
            {
                "id": "polite_announcement",
                "event_type": "object:announcement",
                "is_frame": False,
                "is_dialog_or_alert": False,
                "detail1": 1,
                "expected_priority": 3,
            },
            {
                "id": "other_announcement",
                "event_type": "object:announcement",
                "is_frame": False,
                "is_dialog_or_alert": False,
                "detail1": 3,
                "expected_priority": 4,
            },
            {
                "id": "invalid_entry",
                "event_type": "object:state-changed:invalid-entry",
                "is_frame": False,
                "is_dialog_or_alert": False,
                "detail1": 0,
                "expected_priority": 5,
            },
            {
                "id": "children_changed",
                "event_type": "object:children-changed:add",
                "is_frame": False,
                "is_dialog_or_alert": False,
                "detail1": 0,
                "expected_priority": 6,
            },
            {
                "id": "default_normal",
                "event_type": "object:text-changed:insert",
                "is_frame": False,
                "is_dialog_or_alert": False,
                "detail1": 0,
                "expected_priority": 4,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_get_priority_various_event_types(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test EventManager._get_priority with various event types and conditions."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = case["event_type"]
        mock_event.detail1 = case["detail1"]
        mock_event.source = test_context.Mock()

        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=case["is_frame"]
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_dialog_or_alert",
            return_value=case["is_dialog_or_alert"],
        )

        result = manager._get_priority(mock_event)
        assert result == case["expected_priority"]
        essential_modules["orca.debug"].print_tokens.assert_called()

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "notification_not_ignored",
                "source_role": "notification",
                "event_type": "object:text-changed:insert",
                "source_is_focused": False,
                "manages_descendants": False,
                "detail1": 0,
                "expected_ignore": False,
            },
            {
                "id": "alert_not_ignored",
                "source_role": "alert",
                "event_type": "object:text-changed:insert",
                "source_is_focused": False,
                "manages_descendants": False,
                "detail1": 0,
                "expected_ignore": False,
            },
            {
                "id": "large_text_insertion_ignored",
                "source_role": "text",
                "event_type": "object:text-changed:insert",
                "source_is_focused": False,
                "manages_descendants": False,
                "detail1": 5001,
                "expected_ignore": True,
            },
            {
                "id": "small_text_insertion_not_ignored",
                "source_role": "text",
                "event_type": "object:text-changed:insert",
                "source_is_focused": False,
                "manages_descendants": False,
                "detail1": 4999,
                "expected_ignore": False,
            },
            {
                "id": "focused_source_not_ignored",
                "source_role": "push_button",
                "event_type": "object:state-changed:focused",
                "source_is_focused": True,
                "manages_descendants": False,
                "detail1": 0,
                "expected_ignore": False,
            },
            {
                "id": "focused_managing_descendants_with_detail1",
                "source_role": "push_button",
                "event_type": "object:state-changed:focused",
                "source_is_focused": True,
                "manages_descendants": True,
                "detail1": 1,
                "expected_ignore": False,
            },
            {
                "id": "window_focused_event_ignored",
                "source_role": "window",
                "event_type": "object:state-changed:focused",
                "source_is_focused": False,
                "manages_descendants": False,
                "detail1": 0,
                "expected_ignore": True,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_ignore_event_conditions(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test EventManager._ignore with various event conditions."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        manager._event_history = {}

        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = case["event_type"]
        mock_event.detail1 = case["detail1"]
        mock_event.detail2 = (
            0 if case["event_type"] != "object:text-changed:insert" else case["detail1"]
        )
        mock_event.source = test_context.Mock()
        mock_event.any_data = test_context.Mock()

        mock_app = test_context.Mock()
        focus_mgr = essential_modules["focus_manager_instance"]
        focus_mgr.get_locus_of_focus.return_value = None

        test_context.patch("time.time", return_value=100.0)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window",
            side_effect=lambda obj: (
                case["source_role"] == "window" if obj == mock_event.source else False
            ),
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text",
            side_effect=lambda obj: (
                case["source_role"] == "text" if obj == mock_event.source else False
            ),
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification",
            side_effect=lambda obj: case["source_role"] == "notification"
            if obj == mock_event.source
            else False,
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert",
            side_effect=lambda obj: (
                case["source_role"] == "alert" if obj == mock_event.source else False
            ),
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused",
            side_effect=lambda obj: (
                case["source_is_focused"] if obj == mock_event.source else False
            ),
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_section", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.manages_descendants",
            side_effect=lambda obj: (
                case["manages_descendants"] if obj == mock_event.source else False
            ),
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=mock_app
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_name", return_value="regular-app"
        )

        result = manager._ignore(mock_event)
        assert result is case["expected_ignore"]

    def test_ignore_live_region_text_insertion(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore for live region text insertions."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        manager._event_history = {}

        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:text-changed:insert"
        mock_event.detail1 = 0
        mock_event.detail2 = 100
        mock_event.source = test_context.Mock()
        mock_event.any_data = test_context.Mock()

        mock_app = test_context.Mock()
        focus_mgr = essential_modules["focus_manager_instance"]
        focus_mgr.get_locus_of_focus.return_value = None

        test_context.patch("time.time", return_value=100.0)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_section", return_value=True
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.manages_descendants", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=mock_app
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_name", return_value="regular-app"
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_attribute",
            side_effect=lambda obj, attr: "polite" if attr == "live" else None,
        )

        result = manager._ignore(mock_event)
        assert result is False

    def test_ignore_event_history_filtering(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore with event history spam filtering."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        manager._event_history = {}

        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:test-event"
        mock_event.detail1 = 0
        mock_event.detail2 = 0
        mock_event.source = test_context.Mock()
        mock_event.any_data = test_context.Mock()

        mock_app = test_context.Mock()
        focus_mgr = essential_modules["focus_manager_instance"]
        focus_mgr.get_locus_of_focus.return_value = None

        original_hash = hash

        def mock_hash(obj):
            if obj is mock_app:
                return 12345
            return original_hash(obj)

        test_context.patch("builtins.hash", new=mock_hash)
        test_context.patch("time.time", return_value=100.0)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_section", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.manages_descendants", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=mock_app
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_name", return_value="regular-app"
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_attribute", return_value=None
        )

        # First call should not be ignored
        result1 = manager._ignore(mock_event)
        assert result1 is False
        assert manager._event_history["object:test-event"] == (12345, 100.0)

        # Second call within 0.1s should be ignored
        test_context.patch("time.time", return_value=100.05)
        result2 = manager._ignore(mock_event)
        assert result2 is True

    def test_is_obsoleted_by_identical_events(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._is_obsoleted_by with identical events in queue."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:test-event"
        mock_event.source = test_context.Mock()
        mock_event.detail1 = 1
        mock_event.detail2 = 2
        mock_event.any_data = "test_data"

        identical_event = test_context.Mock(spec=Atspi.Event)
        identical_event.type = "object:test-event"
        identical_event.source = mock_event.source
        identical_event.detail1 = 1
        identical_event.detail2 = 2
        identical_event.any_data = "test_data"

        queue_data = [(4, 1, identical_event)]
        with manager._event_queue.mutex:
            manager._event_queue.queue = queue_data

        result = manager._is_obsoleted_by(mock_event)
        assert result == identical_event

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "active_descendant_obsoletes",
                "existing_type": "object:active-descendant-changed",
                "existing_source_same": True,
                "new_type": "object:active-descendant-changed",
                "should_obsolete": True,
            },
            {
                "id": "state_changed_obsoletes",
                "existing_type": "object:state-changed:focused",
                "existing_source_same": True,
                "new_type": "object:state-changed:focused",
                "should_obsolete": True,
            },
            {
                "id": "caret_moved_obsoletes",
                "existing_type": "object:text-caret-moved",
                "existing_source_same": True,
                "new_type": "object:text-caret-moved",
                "should_obsolete": True,
            },
            {
                "id": "window_activate_obsoletes",
                "existing_type": "window:activate",
                "existing_source_same": True,
                "new_type": "window:activate",
                "should_obsolete": True,
            },
            {
                "id": "non_skippable_does_not_obsolete",
                "existing_type": "object:other-event",
                "existing_source_same": True,
                "new_type": "object:other-event",
                "should_obsolete": False,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_is_obsoleted_by_same_type_and_object(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test EventManager._is_obsoleted_by with same type and object conditions."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_source = test_context.Mock()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = case["new_type"]
        mock_event.source = mock_source
        mock_event.detail1 = 0
        mock_event.detail2 = 0
        mock_event.any_data = None

        existing_event = test_context.Mock(spec=Atspi.Event)
        existing_event.type = case["existing_type"]
        existing_event.source = mock_source if case["existing_source_same"] else test_context.Mock()
        existing_event.detail1 = (
            0 if case["should_obsolete"] else 1
        )  # Make it different for non-obsoleting case
        existing_event.detail2 = 0
        existing_event.any_data = None

        queue_data = [(4, 1, existing_event)]
        with manager._event_queue.mutex:
            manager._event_queue.queue = queue_data

        result = manager._is_obsoleted_by(mock_event)
        if case["should_obsolete"]:
            assert result == existing_event
        else:
            assert result is None

    def test_is_obsoleted_by_sibling_events(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._is_obsoleted_by with sibling event conditions."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        parent_mock = test_context.Mock()
        mock_source = test_context.Mock()
        sibling_source = test_context.Mock()

        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:state-changed:focused"
        mock_event.source = mock_source
        mock_event.detail1 = 1
        mock_event.detail2 = 2
        mock_event.any_data = "test_data"

        sibling_event = test_context.Mock(spec=Atspi.Event)
        sibling_event.type = "object:state-changed:focused"
        sibling_event.source = sibling_source
        sibling_event.detail1 = 1
        sibling_event.detail2 = 2
        sibling_event.any_data = "test_data"

        test_context.patch(
            "orca.event_manager.AXObject.get_parent",
            side_effect=lambda obj: parent_mock
            if obj in [mock_source, sibling_source]
            else test_context.Mock(),
        )

        queue_data = [(4, 1, sibling_event)]
        with manager._event_queue.mutex:
            manager._event_queue.queue = queue_data

        result = manager._is_obsoleted_by(mock_event)
        assert result == sibling_event

    def test_is_obsoleted_by_window_events(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._is_obsoleted_by with window event conditions."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_source = test_context.Mock()

        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "window:activate"
        mock_event.source = mock_source
        mock_event.detail1 = 0
        mock_event.detail2 = 0
        mock_event.any_data = None

        existing_event = test_context.Mock(spec=Atspi.Event)
        existing_event.type = "window:activate"
        existing_event.source = mock_source
        existing_event.detail1 = 0
        existing_event.detail2 = 0
        existing_event.any_data = None

        queue_data = [(4, 1, existing_event)]
        with manager._event_queue.mutex:
            manager._event_queue.queue = queue_data

        result = manager._is_obsoleted_by(mock_event)
        assert result == existing_event

    def test_focus_conditions_in_ignore(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore focus-related conditions."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        manager._event_history = {}

        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:test-event"
        mock_event.detail1 = 0
        mock_event.detail2 = 0
        mock_event.source = test_context.Mock()
        mock_event.any_data = test_context.Mock()

        focus_mgr = essential_modules["focus_manager_instance"]

        focus_mgr.get_locus_of_focus.return_value = mock_event.source
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )

        result = manager._ignore(mock_event)
        assert result is False

        focus_mgr.get_locus_of_focus.return_value = mock_event.any_data
        result = manager._ignore(mock_event)
        assert result is False

    def test_ignore_selected_source_not_ignored(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._ignore does not ignore events from selected sources."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        manager._event_history = {}

        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:test-event"
        mock_event.detail1 = 0
        mock_event.detail2 = 0
        mock_event.source = test_context.Mock()
        mock_event.any_data = test_context.Mock()

        mock_app = test_context.Mock()
        focus_mgr = essential_modules["focus_manager_instance"]
        focus_mgr.get_locus_of_focus.return_value = None

        test_context.patch("time.time", return_value=100.0)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_window", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_frame", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_text", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_notification", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_alert", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_selected",
            return_value=True,  # Selected source should not be ignored
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_focused", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.is_section", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.manages_descendants", return_value=False
        )
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=mock_app
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_name", return_value="regular-app"
        )
        test_context.patch(
            "orca.event_manager.AXObject.get_attribute", return_value=None
        )

        result = manager._ignore(mock_event)
        assert result is False

    def test_queue_println(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._queue_println."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "test:event"
        debug_mock = essential_modules["orca.debug"]
        debug_mock.debugLevel = 10
        manager._queue_println(mock_event)
        manager._queue_println(mock_event, is_enqueue=False)

    def test_enqueue_object_event_ignored(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._enqueue_object_event for ignored events."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = False
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:text-changed:insert"
        manager._enqueue_object_event(mock_event)
        assert manager._event_queue.empty()

    def test_on_no_focus(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._on_no_focus."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_get_focus_mgr = test_context.Mock()
        test_context.patch(
            "orca.event_manager.focus_manager.get_manager", new=mock_get_focus_mgr
        )
        mock_get_script_mgr = test_context.Mock()
        test_context.patch(
            "orca.event_manager.script_manager.get_manager", new=mock_get_script_mgr
        )
        mock_braille = test_context.Mock()
        test_context.patch("orca.event_manager.braille", new=mock_braille)
        mock_focus_mgr = test_context.Mock()
        mock_script_mgr = test_context.Mock()
        mock_get_focus_mgr.return_value = mock_focus_mgr
        mock_get_script_mgr.return_value = mock_script_mgr

        mock_focus_mgr.focus_and_window_are_unknown.return_value = True
        result = manager._on_no_focus()
        assert result is False

        mock_focus_mgr.focus_and_window_are_unknown.return_value = False
        mock_script_mgr.get_active_script.return_value = None
        result = manager._on_no_focus()
        assert result is False
        mock_script_mgr.set_active_script.assert_called_once()
        mock_braille.disableBraille.assert_called_once()

    def test_dequeue_object_event_empty_queue(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._dequeue_object_event with empty queue."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        result = manager._dequeue_object_event()
        assert result is False
        assert manager._gidle_id == 0

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "first_registration",
                "operation": "register",
                "initial_count": 0,
                "expected_count": 1,
                "should_call_listener": True,
            },
            {
                "id": "duplicate_registration",
                "operation": "register",
                "initial_count": 1,
                "expected_count": 2,
                "should_call_listener": False,
            },
            {
                "id": "partial_deregistration",
                "operation": "deregister",
                "initial_count": 2,
                "expected_count": 1,
                "should_call_listener": False,
            },
            {
                "id": "final_deregistration",
                "operation": "deregister",
                "initial_count": 1,
                "expected_count": 0,
                "should_call_listener": True,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_listener_registration(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test EventManager register/deregister listener operations."""
        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._listener = test_context.Mock()
        event_type = "object:text-changed:insert"

        if case["initial_count"] > 0:
            manager._script_listener_counts[event_type] = case["initial_count"]

        if case["operation"] == "register":
            manager.register_listener(event_type)
            if case["should_call_listener"]:
                manager._listener.register.assert_called_once_with(event_type)
            else:
                manager._listener.register.assert_not_called()
        else:  # deregister
            manager.deregister_listener(event_type)
            if case["should_call_listener"]:
                manager._listener.deregister.assert_called_once_with(event_type)
            else:
                manager._listener.deregister.assert_not_called()

        if case["expected_count"] == 0:
            assert event_type not in manager._script_listener_counts
        else:
            assert manager._script_listener_counts[event_type] == case["expected_count"]

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "register_script_listeners", "operation": "register"},
            {"id": "deregister_script_listeners", "operation": "deregister"},
        ],
        ids=lambda case: case["id"],
    )
    def test_script_listeners_operations(self, test_context: OrcaTestContext, case: dict) -> None:
        """Test EventManager register/deregister script listeners operations."""
        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_listener_method = test_context.Mock()
        method_name = f"{case['operation']}_listener"
        test_context.patch_object(manager, method_name, new=mock_listener_method)

        mock_script = test_context.Mock()
        mock_script.listeners = {
            "object:text-changed:insert": test_context.Mock(),
            "object:state-changed:focused": test_context.Mock(),
        }

        getattr(manager, f"{case['operation']}_script_listeners")(mock_script)

        expected_calls = [
            call("object:text-changed:insert"),
            call("object:state-changed:focused"),
        ]
        mock_listener_method.assert_has_calls(expected_calls, any_order=True)

    def test_get_script_for_event_focus(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._get_script_for_event for focused events."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.source = test_context.Mock()
        mock_event.type = "object:state-changed:focused"
        active_script = test_context.Mock()
        mock_get_focus_mgr = test_context.Mock()
        test_context.patch(
            "orca.event_manager.focus_manager.get_manager", new=mock_get_focus_mgr
        )
        mock_focus_mgr = test_context.Mock()
        mock_get_focus_mgr.return_value = mock_focus_mgr
        mock_focus_mgr.get_locus_of_focus.return_value = mock_event.source
        result = manager._get_script_for_event(mock_event, active_script)
        assert result == active_script

    def test_get_script_for_event_mouse(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._get_script_for_event for mouse events."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "mouse:button:1p"
        mock_event.source = test_context.Mock()
        expected_script = test_context.Mock()
        mock_mouse_event = test_context.Mock()
        mock_mouse_event.app = test_context.Mock()
        mock_mouse_event.window = test_context.Mock()
        mock_get_script_mgr = test_context.Mock()
        test_context.patch(
            "orca.event_manager.script_manager.get_manager", new=mock_get_script_mgr
        )
        test_context.patch(
            "orca.event_manager.input_event.MouseButtonEvent", return_value=mock_mouse_event
        )
        mock_script_mgr = test_context.Mock()
        mock_get_script_mgr.return_value = mock_script_mgr
        mock_script_mgr.get_script.return_value = expected_script
        result = manager._get_script_for_event(mock_event)
        assert result == expected_script

    def test_get_script_for_event_defunct_app(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._get_script_for_event with defunct application."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:text-changed:insert"
        mock_event.source = test_context.Mock()
        mock_app = test_context.Mock()
        ax_utilities = essential_modules["orca.ax_utilities"]
        ax_utilities.get_application.return_value = mock_app
        ax_utilities.is_defunct.return_value = True
        result = manager._get_script_for_event(mock_event)
        assert result is None

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "focus_claimed",
                "event_type": "object:state-changed:focused",
                "detail1": 1,
                "expected": True,
                "reason_contains": "claimed focus",
            },
            {
                "id": "menu_selection",
                "event_type": "object:state-changed:selected",
                "detail1": 1,
                "expected": True,
                "reason_contains": "Selection change",
            },
            {
                "id": "modal_panel",
                "event_type": "object:state-changed:showing",
                "detail1": 0,
                "expected": True,
                "reason_contains": "Modal panel",
            },
            {
                "id": "no_activation",
                "event_type": "object:text-changed:insert",
                "detail1": 0,
                "expected": False,
                "reason_contains": "No reason",
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_is_activatable_event(self, test_context: OrcaTestContext, case: dict) -> None:
        """Test EventManager._is_activatable_event."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = case["event_type"]
        mock_event.detail1 = case["detail1"]
        mock_event.source = test_context.Mock()
        mock_script = test_context.Mock()
        mock_script.is_activatable_event.return_value = True
        mock_script.force_script_activation.return_value = False
        ax_utilities = essential_modules["orca.ax_utilities"]
        if case["event_type"] == "object:state-changed:selected":
            ax_utilities.is_menu.return_value = True
            ax_utilities.is_focusable.return_value = True
        elif case["event_type"] == "object:state-changed:showing":
            ax_utilities.is_panel.return_value = True
            ax_utilities.is_modal.return_value = True
        else:
            ax_utilities.is_frame.return_value = case["event_type"] == "object:state-changed:active"
        focus_mgr = essential_modules["focus_manager_instance"]
        if case["event_type"] == "window:activate":
            focus_mgr.get_active_window.return_value = mock_event.source
        else:
            focus_mgr.get_active_window.return_value = test_context.Mock()
        result, reason = manager._is_activatable_event(mock_event, mock_script)
        assert result == case["expected"]
        assert case["reason_contains"].lower() in reason.lower()

    def test_is_activatable_event_no_source(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._is_activatable_event with no event source."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.source = None
        result, reason = manager._is_activatable_event(mock_event)
        assert result is False
        assert "event.source" in reason

    def test_is_activatable_event_script_forces(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._is_activatable_event when script forces activation."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:text-changed:insert"
        mock_event.source = test_context.Mock()
        mock_script = test_context.Mock()
        mock_script.is_activatable_event.return_value = True
        mock_script.force_script_activation.return_value = True
        result, reason = manager._is_activatable_event(mock_event, mock_script)
        assert result is True
        assert "insists" in reason

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "active_script",
                "event_script_is_active": True,
                "present_if_inactive": False,
                "is_progress_bar": False,
                "verbosity_all": False,
                "expected": True,
            },
            {
                "id": "inactive_but_presents",
                "event_script_is_active": False,
                "present_if_inactive": True,
                "is_progress_bar": False,
                "verbosity_all": False,
                "expected": True,
            },
            {
                "id": "progress_bar_all",
                "event_script_is_active": False,
                "present_if_inactive": False,
                "is_progress_bar": True,
                "verbosity_all": True,
                "expected": True,
            },
            {
                "id": "no_reason_to_process",
                "event_script_is_active": False,
                "present_if_inactive": False,
                "is_progress_bar": False,
                "verbosity_all": False,
                "expected": False,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_should_process_event(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test EventManager._should_process_event."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "test:event"
        mock_event.source = test_context.Mock()
        event_script = test_context.Mock()
        event_script.present_if_inactive = case["present_if_inactive"]
        if case["event_script_is_active"]:
            active_script = event_script
        else:
            active_script = test_context.Mock()
        mock_settings = test_context.Mock()
        test_context.patch("orca.event_manager.settings", new=mock_settings)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_progress_bar", return_value=case["is_progress_bar"]
        )
        if case["verbosity_all"]:
            mock_settings.progressBarVerbosity = mock_settings.PROGRESS_BAR_ALL
        else:
            mock_settings.progressBarVerbosity = 1
        result = manager._should_process_event(mock_event, event_script, active_script)
        assert result == case["expected"]

    def test_process_object_event_obsoleted(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._process_object_event with obsoleted event."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_is_obsoleted = test_context.Mock(return_value=test_context.Mock())
        test_context.patch_object(manager, "_is_obsoleted_by", new=mock_is_obsoleted)
        manager._process_object_event(mock_event)

    def test_process_object_event_dead_source(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._process_object_event with dead event source."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "window:deactivate"
        mock_event.source = test_context.Mock()
        mock_is_obsoleted = test_context.Mock(return_value=None)
        test_context.patch_object(manager, "_is_obsoleted_by", new=mock_is_obsoleted)
        test_context.patch("orca.event_manager.AXObject.is_dead", return_value=True)
        test_context.patch(
            "orca.event_manager.AXUtilities.is_defunct", return_value=False
        )
        mock_get_focus_mgr = test_context.Mock()
        test_context.patch(
            "orca.event_manager.focus_manager.get_manager", new=mock_get_focus_mgr
        )
        mock_get_script_mgr = test_context.Mock()
        test_context.patch(
            "orca.event_manager.script_manager.get_manager", new=mock_get_script_mgr
        )
        mock_focus_mgr = test_context.Mock()
        mock_script_mgr = test_context.Mock()
        mock_get_focus_mgr.return_value = mock_focus_mgr
        mock_get_script_mgr.return_value = mock_script_mgr
        mock_focus_mgr.get_active_window.return_value = mock_event.source
        manager._process_object_event(mock_event)

        mock_focus_mgr.clear_state.assert_called_once()
        mock_script_mgr.set_active_script.assert_called_once_with(
            None, "Active window is dead or defunct"
        )

    def test_process_object_event_no_listener(self, test_context: OrcaTestContext) -> None:
        """Test EventManager._process_object_event with no matching listener."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:unknown-event"
        mock_event.source = test_context.Mock()
        mock_is_obsoleted = test_context.Mock(return_value=None)
        test_context.patch_object(manager, "_is_obsoleted_by", new=mock_is_obsoleted)
        ax_object = essential_modules["orca.ax_object"]
        ax_utilities = essential_modules["orca.ax_utilities"]
        ax_object.is_dead.return_value = False
        ax_utilities.is_defunct.return_value = False
        ax_utilities.is_iconified.return_value = False
        mock_script = test_context.Mock()
        mock_script.listeners = {}
        script_mgr = essential_modules["script_manager_instance"]
        script_mgr.get_active_script.return_value = mock_script
        mock_get_script = test_context.Mock(return_value=mock_script)
        mock_is_activatable = test_context.Mock(return_value=(False, "test"))
        mock_should_process = test_context.Mock(return_value=True)
        test_context.patch_object(manager, "_get_script_for_event", new=mock_get_script)
        test_context.patch_object(manager, "_is_activatable_event", new=mock_is_activatable)
        test_context.patch_object(manager, "_should_process_event", new=mock_should_process)
        manager._process_object_event(mock_event)

    def test_get_manager(self, test_context: OrcaTestContext) -> None:
        """Test event_manager.get_manager."""

        self._setup_dependencies(test_context)
        from orca import event_manager

        manager1 = event_manager.get_manager()
        manager2 = event_manager.get_manager()

        assert manager1 is manager2
        assert isinstance(manager1, event_manager.EventManager)

    def test_enqueue_with_comprehensive_gidle_management(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test EventManager._enqueue_object_event with comprehensive GLib idle management."""

        self._setup_dependencies(test_context)
        from orca.event_manager import EventManager

        manager = EventManager()
        manager._active = True
        manager._paused = False
        manager._gidle_id = 0  # No existing idle handler
        mock_event = test_context.Mock(spec=Atspi.Event)
        mock_event.type = "object:text-changed:insert"
        mock_event.source = test_context.Mock()
        mock_app = test_context.Mock()
        mock_script = test_context.Mock()
        mock_script.event_cache = {}

        mock_ignore = test_context.Mock(return_value=False)
        test_context.patch_object(manager, "_ignore", new=mock_ignore)
        test_context.patch(
            "orca.event_manager.AXUtilities.get_application", return_value=mock_app
        )
        mock_get_script_mgr = test_context.Mock()
        test_context.patch(
            "orca.event_manager.script_manager.get_manager", new=mock_get_script_mgr
        )
        mock_idle_add = test_context.Mock(return_value=456)
        test_context.patch("orca.event_manager.GLib.idle_add", new=mock_idle_add)
        test_context.patch("time.time", return_value=200.0)
        mock_script_mgr = test_context.Mock()
        mock_get_script_mgr.return_value = mock_script_mgr
        mock_script_mgr.get_script.return_value = mock_script
        manager._enqueue_object_event(mock_event)

        assert not manager._event_queue.empty()
        assert mock_event.type in mock_script.event_cache
        mock_idle_add.assert_called_once_with(manager._dequeue_object_event)
        assert manager._gidle_id == 456

        mock_idle_add.reset_mock()
        manager._gidle_id = 456
        test_context.patch("time.time", return_value=200.1)
        manager._enqueue_object_event(mock_event)

        mock_idle_add.assert_not_called()
        assert manager._gidle_id == 456



# Orca
#
# Copyright 2024 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=no-name-in-module

"""Provides debugging tools."""

# This has to be the first non-docstring line in the module to make linters happy.
from __future__ import annotations

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2024 Igalia, S.L." \
                "Copyright (c) 2024 GNOME Foundation Inc."
__license__   = "LGPL"

import faulthandler
import os
import subprocess
import time
from typing import Generator, TYPE_CHECKING

import gi
gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from . import cmdnames
from . import debug
from . import focus_manager
from . import input_event
from . import keybindings
from . import messages
from . import orca_platform
from . import settings_manager
from .ax_object import AXObject
from .ax_utilities import AXUtilities
from .ax_utilities_debugging import AXUtilitiesDebugging

if TYPE_CHECKING:
    from .scripts import default

class DebuggingToolsManager:
    """Provides debugging tools."""

    def __init__(self) -> None:
        self._handlers: dict[str, input_event.InputEventHandler] = self.get_handlers(True)
        self._bindings: keybindings.KeyBindings = keybindings.KeyBindings()

        if debug.debugFile and os.path.exists(debug.debugFile.name):
            faulthandler.enable(file=debug.debugFile, all_threads=True)
        else:
            faulthandler.enable(all_threads=False)

    def get_bindings(
        self, refresh: bool = False, is_desktop: bool = True
    ) -> keybindings.KeyBindings:
        """Returns the debugging-tools-manager keybindings."""

        if refresh:
            msg = f"DEBUGGING TOOLS MANAGER: Refreshing bindings. Is desktop: {is_desktop}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._bindings.remove_key_grabs("DEBUGGING TOOLS MANAGER: Refreshing bindings.")
            self._setup_bindings()
        elif self._bindings.is_empty():
            self._setup_bindings()

        return self._bindings

    def get_handlers(self, refresh: bool = False) -> dict[str, input_event.InputEventHandler]:
        """Returns the debugging-tools-manager handlers."""

        if refresh:
            msg = "DEBUGGING TOOLS MANAGER: Refreshing handlers."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._setup_handlers()

        return self._handlers

    def _setup_handlers(self) -> None:
        """Sets up and returns the debugging-tools-manager input event handlers."""

        self._handlers = {}

        self._handlers["cycleDebugLevelHandler"] = \
            input_event.InputEventHandler(
                self._cycle_debug_level,
                cmdnames.DEBUG_CYCLE_LEVEL)

        self._handlers["clear_atspi_app_cache"] = \
            input_event.InputEventHandler(
                self._clear_atspi_app_cache,
                cmdnames.DEBUG_CLEAR_ATSPI_CACHE_FOR_APPLICATION)

        self._handlers["capture_snapshot"] = \
            input_event.InputEventHandler(
                self._capture_snapshot,
                cmdnames.DEBUG_CAPTURE_SNAPSHOT)

    def _setup_bindings(self) -> None:
        """Sets up and returns the debugging-tools-manager key bindings."""

        self._bindings = keybindings.KeyBindings()

        self._bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["cycleDebugLevelHandler"],
                1,
                True))

        self._bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["clear_atspi_app_cache"],
                1,
                True))

        self._bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["capture_snapshot"],
                1,
                True))

        # This pulls in the user's overrides to alternative keys.
        self._bindings = settings_manager.get_manager().override_key_bindings(
            self._handlers, self._bindings, False)

    def _cycle_debug_level(
        self, script: default.Script, _event: input_event.InputEvent | None = None
    ) -> bool:
        """Cycles through the existing debug levels."""

        levels = {
            debug.LEVEL_ALL: "all",
            debug.LEVEL_INFO: "info",
            debug.LEVEL_WARNING: "warning",
            debug.LEVEL_SEVERE: "severe",
            debug.LEVEL_OFF: "off",
        }

        keys = list(levels.keys())
        next_level = keys.index(debug.debugLevel) + 1
        if next_level == len(keys):
            next_level = 0

        level = keys[next_level]
        brief = levels.get(level)
        debug.debugLevel = level
        script.present_message(f"Debug level {brief}.", brief)
        return True

    def _clear_atspi_app_cache(
        self, script: default.Script, _event: input_event.InputEvent | None = None
    ) -> bool:
        """Clears the AT-SPI cache for the current application."""

        _mode, obj = focus_manager.get_manager().get_active_mode_and_object_of_interest()
        if obj is None:
            msg = "DEBUGGING TOOLS MANAGER: Cannot clear cache on null object of interest."
            debug.print_message(debug.debugLevel, msg, True)
            script.present_message(messages.DEBUG_CLEAR_CACHE_FAILED)
            return True

        app = AXUtilities.get_application(obj)
        if app is None:
            msg = "DEBUGGING TOOLS MANAGER: Cannot clear cache on null application."
            debug.print_message(debug.debugLevel, msg, True)
            script.present_message(messages.DEBUG_CLEAR_CACHE_FAILED)
            return True

        script.present_message(messages.DEBUG_CLEAR_CACHE)
        AXObject.clear_cache(app, recursive=True, reason="User request.")
        return True

    def _capture_snapshot(
        self, script: default.Script, _event: input_event.InputEvent | None = None
    ) -> bool:
        """Clears the AT-SPI cache for the current application."""

        script.present_message(messages.DEBUG_CAPTURE_SNAPSHOT_START)

        old_level = debug.debugLevel
        debug.debugLevel = debug.LEVEL_SEVERE
        debug.print_message(debug.debugLevel, "DEBUGGING SNAPSHOT STARTING", True)
        self.print_running_applications()

        manager = settings_manager.get_manager()
        info = AXUtilitiesDebugging.as_string(manager.get_overridden_settings_for_debugging())
        msg = f"OVERRIDDEN SETTINGS: {info}"
        debug.print_message(debug.debugLevel, msg, True)

        info = AXUtilitiesDebugging.as_string(manager.customized_settings)
        msg = f"CUSTOMIZED SETTINGS: {info}"
        debug.print_message(debug.debugLevel, msg, True)

        debug.print_message(debug.debugLevel, "DEBUGGING SNAPSHOT FINISHED", True)
        script.present_message(messages.DEBUG_CAPTURE_SNAPSHOT_END)
        debug.debugLevel = old_level
        return True

    def _get_running_applications_as_string_iter(
        self, is_command_line: bool
    ) -> Generator[str, None, None]:
        """Generator providing strings with basic details about the running accessible apps."""

        applications = AXUtilities.get_all_applications(is_debug=True)
        msg = f"Desktop has {len(applications)} app(s):"
        if not is_command_line:
            msg = f"DEBUGGING TOOLS MANAGER: {msg}"
        yield msg

        for i, app in enumerate(applications):
            pid = AXUtilities.get_process_id(app)
            if AXUtilities.is_application_unresponsive(app):
                name = "[UNRESPONSIVE]"
            else:
                name = AXObject.get_name(app) or "[DEAD]"
            try:
                cmdline = subprocess.getoutput(f"cat /proc/{pid}/cmdline")
            except subprocess.SubprocessError as error:
                cmdline = f"EXCEPTION: {error}"
            else:
                cmdline = cmdline.replace("\x00", " ")
            if is_command_line:
                prefix = f"{time.strftime('%H:%M:%S', time.localtime()):<12}"
            else:
                prefix = f"{i+1:3}."

            msg = f"{prefix} pid: {pid:<10} {name:<25} {cmdline}"
            yield msg

    def print_running_applications(
        self, force: bool = False, is_command_line: bool = False
    ) -> None:
        """Prints basic details about the running accessible applications."""

        if force:
            level = debug.LEVEL_SEVERE
        else:
            level = debug.LEVEL_INFO

        if level < debug.debugLevel and not is_command_line:
            return

        for app_string in self._get_running_applications_as_string_iter(is_command_line):
            if is_command_line:
                print(app_string)
            else:
                debug.print_message(level, app_string, True)

    def print_session_details(self, is_command_line: bool = False) -> None:
        """Prints basic details about the current session."""

        msg = f"Orca version {orca_platform.version}"
        if orca_platform.revision:
            msg += f" (rev {orca_platform.revision})"

        atspi_version = Atspi.get_version()
        msg += f", AT-SPI2 version: {atspi_version[0]}.{atspi_version[1]}.{atspi_version[2]}"
        session_type = os.environ.get("XDG_SESSION_TYPE") or ""
        session_desktop = os.environ.get("XDG_SESSION_DESKTOP") or ""
        session = f"{session_type} {session_desktop}".strip()
        if session:
            msg += f", Session: {session}"

        if is_command_line:
            print(msg)
        else:
            msg = f"DEBUGGING TOOLS MANAGER: {msg}"
            debug.print_message(debug.LEVEL_INFO, msg, True)

_manager: DebuggingToolsManager = DebuggingToolsManager()

def get_manager() -> DebuggingToolsManager:
    """Returns the debugging tools manager."""

    return _manager



# Unit tests for bypass_mode_manager.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=import-outside-toplevel
# pylint: disable=protected-access
# pylint: disable=no-member

"""Unit tests for bypass_mode_manager.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext
    from unittest.mock import MagicMock

@pytest.mark.unit
class TestBypassModeManager:
    """Test BypassModeManager class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, MagicMock]:
        """Returns dependencies for bypass_mode_manager module testing."""

        additional_modules = ["orca.orca_modifier_manager"]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)

        debug_mock = essential_modules["orca.debug"]
        debug_mock.print_message = test_context.Mock()
        debug_mock.LEVEL_INFO = 800

        cmdnames_mock = essential_modules["orca.cmdnames"]
        cmdnames_mock.BYPASS_MODE_TOGGLE = "Toggle all Orca command keys"

        input_event_mock = essential_modules["orca.input_event"]
        input_event_handler_mock = test_context.Mock()
        input_event_mock.InputEventHandler = test_context.Mock(
            return_value=input_event_handler_mock
        )

        keybindings_mock = essential_modules["orca.keybindings"]
        key_bindings_instance = test_context.Mock()
        key_bindings_instance.is_empty = test_context.Mock(return_value=False)
        key_bindings_instance.key_bindings = []
        key_bindings_instance.add = test_context.Mock(return_value=None)
        keybindings_mock.KeyBindings = test_context.Mock(return_value=key_bindings_instance)
        keybindings_mock.KeyBinding = test_context.Mock()
        keybindings_mock.DEFAULT_MODIFIER_MASK = 1
        keybindings_mock.ALT_MODIFIER_MASK = 2

        messages_mock = essential_modules["orca.messages"]
        messages_mock.BYPASS_MODE_ENABLED = "Orca command keys off."
        messages_mock.BYPASS_MODE_DISABLED = "Orca command keys on."

        orca_modifier_manager_mock = essential_modules["orca.orca_modifier_manager"]
        modifier_manager_instance = test_context.Mock()
        modifier_manager_instance.unset_orca_modifiers = test_context.Mock()
        modifier_manager_instance.refresh_orca_modifiers = test_context.Mock()
        orca_modifier_manager_mock.get_manager = test_context.Mock(
            return_value=modifier_manager_instance
        )

        settings_manager_mock = essential_modules["orca.settings_manager"]
        settings_manager_instance = test_context.Mock()
        override_method = test_context.Mock(return_value=key_bindings_instance)
        settings_manager_instance.override_key_bindings = override_method
        settings_manager_mock.get_manager = test_context.Mock(
            return_value=settings_manager_instance
        )

        essential_modules["input_event_handler"] = input_event_handler_mock
        essential_modules["key_bindings_instance"] = key_bindings_instance
        essential_modules["modifier_manager_instance"] = modifier_manager_instance
        essential_modules["settings_manager_instance"] = settings_manager_instance

        return essential_modules

    def test_init(self, test_context: OrcaTestContext) -> None:
        """Test BypassModeManager initialization."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        assert manager._is_active is False
        assert manager._handlers is not None
        assert manager._bindings is not None
        essential_modules["orca.keybindings"].KeyBindings.assert_called()

    @pytest.mark.parametrize(
        "method_name,is_desktop,expected_log_msg",
        [
            pytest.param(
                "get_bindings",
                True,
                "BYPASS MODE MANAGER: Refreshing bindings. Is desktop: True",
                id="bindings_desktop",
            ),
            pytest.param(
                "get_bindings",
                False,
                "BYPASS MODE MANAGER: Refreshing bindings. Is desktop: False",
                id="bindings_not_desktop",
            ),
            pytest.param(
                "get_handlers",
                None,
                "BYPASS MODE MANAGER: Refreshing handlers.",
                id="handlers_refresh",
            ),
        ],
    )
    def test_get_bindings_and_handlers_refresh(
        self,
        test_context: OrcaTestContext,
        method_name: str,
        is_desktop: bool | None,
        expected_log_msg: str,
    ) -> None:
        """Test get_bindings and get_handlers with refresh=True."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        essential_modules["orca.debug"].print_message.reset_mock()

        if method_name == "get_bindings":
            bindings_result = manager.get_bindings(refresh=True, is_desktop=is_desktop or False)
            assert bindings_result is not None
        else:
            handlers_result = manager.get_handlers(refresh=True)
            assert handlers_result is not None
        essential_modules["orca.debug"].print_message.assert_called_with(
            essential_modules["orca.debug"].LEVEL_INFO,
            expected_log_msg,
            True,
        )

    def test_setup_handlers(self, test_context: OrcaTestContext) -> None:
        """Test _setup_handlers method."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        manager._setup_handlers()
        assert "bypass_mode_toggle" in manager._handlers
        expected_handler = essential_modules["input_event_handler"]
        assert manager._handlers["bypass_mode_toggle"] == expected_handler
        input_event = essential_modules["orca.input_event"]
        input_event.InputEventHandler.assert_called_with(
            manager.toggle_enabled, "Toggle all Orca command keys"
        )

    @pytest.mark.parametrize(
        "default_mask,alt_mask",
        [
            pytest.param(1, 2, id="original_config"),
            pytest.param(8, 16, id="custom_config"),
        ],
    )
    def test_setup_bindings(
        self, test_context: OrcaTestContext, default_mask: int, alt_mask: int
    ) -> None:
        """Test _setup_bindings method with different key binding configurations."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)

        keybindings_mock = essential_modules["orca.keybindings"]
        keybindings_mock.DEFAULT_MODIFIER_MASK = default_mask
        keybindings_mock.ALT_MODIFIER_MASK = alt_mask
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        keybindings_mock.KeyBinding.reset_mock()
        manager._setup_bindings()

        essential_modules["orca.keybindings"].KeyBinding.assert_called_with(
            "BackSpace",
            default_mask,
            alt_mask,
            manager._handlers["bypass_mode_toggle"],
            1,
            True,
        )

        key_bindings_instance = essential_modules["key_bindings_instance"]
        key_bindings_instance.add.assert_called()
        settings_manager_instance = essential_modules["settings_manager_instance"]
        settings_manager_instance.override_key_bindings.assert_called_with(
            manager._handlers, key_bindings_instance, False
        )

    @pytest.mark.parametrize(
        "initial_active,has_event,expected_active,expected_message,"
        "expected_grabs_call,expected_modifier_call",
        [
            (
                False,
                True,
                True,
                "Orca command keys off.",
                "remove_key_grabs",
                "unset_orca_modifiers",
            ),
            (
                False,
                False,
                True,
                None,
                "remove_key_grabs",
                "unset_orca_modifiers",
            ),
            (
                True,
                True,
                False,
                "Orca command keys on.",
                "add_key_grabs",
                "refresh_orca_modifiers",
            ),
            (
                True,
                False,
                False,
                None,
                "add_key_grabs",
                "refresh_orca_modifiers",
            ),
        ],
    )
    def test_toggle_enabled(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        test_context: OrcaTestContext,
        initial_active: bool,
        has_event: bool,
        expected_active: bool,
        expected_message: str | None,
        expected_grabs_call: str,
        expected_modifier_call: str,
    ) -> None:
        """Test toggle_enabled with various initial states and event conditions."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)

        if not initial_active:
            mock_binding1 = test_context.Mock()
            mock_binding2 = test_context.Mock()
            key_bindings = [mock_binding1, mock_binding2]
        else:
            key_bindings = []

        key_bindings_instance = essential_modules["key_bindings_instance"]
        key_bindings_instance.key_bindings = key_bindings
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        manager._is_active = initial_active
        mock_script = test_context.Mock()
        mock_event = test_context.Mock() if has_event else None

        result = manager.toggle_enabled(mock_script, mock_event)
        assert result is True
        assert manager._is_active is expected_active

        if expected_message:
            mock_script.present_message.assert_called_with(expected_message)
        else:
            mock_script.present_message.assert_not_called()

        grabs_method = getattr(mock_script, expected_grabs_call)
        if expected_active:
            grabs_method.assert_called_with("bypass mode enabled")
        else:
            grabs_method.assert_called_with("bypass mode disabled")

        modifier_manager = essential_modules["modifier_manager_instance"]
        modifier_method = getattr(modifier_manager, expected_modifier_call)
        if expected_active:
            modifier_method.assert_called_with("bypass mode enabled")
        else:
            modifier_method.assert_called_with("bypass mode disabled")

        if not initial_active and key_bindings:
            mock_script.key_bindings.add.assert_any_call(key_bindings[0], include_grabs=True)
            mock_script.key_bindings.add.assert_any_call(key_bindings[1], include_grabs=True)

    def test_toggle_enabled_multiple_times(self, test_context: OrcaTestContext) -> None:
        """Test toggle_enabled works correctly when called multiple times."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)

        key_bindings_instance = essential_modules["key_bindings_instance"]
        key_bindings_instance.key_bindings = []
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        mock_script = test_context.Mock()

        assert manager.is_active() is False

        result1 = manager.toggle_enabled(mock_script, None)
        assert result1 is True
        assert manager.is_active() is True

        result2 = manager.toggle_enabled(mock_script, None)
        assert result2 is True
        assert manager.is_active() is False

        result3 = manager.toggle_enabled(mock_script, None)
        assert result3 is True
        assert manager.is_active() is True

    def test_bindings_with_key_override(self, test_context: OrcaTestContext) -> None:
        """Test that bindings are properly configured with key overrides."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)

        mock_overridden_bindings = test_context.Mock()
        settings_manager_instance = essential_modules["settings_manager_instance"]
        settings_manager_instance.override_key_bindings.return_value = mock_overridden_bindings
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        bindings = manager.get_bindings(refresh=True)
        assert bindings == mock_overridden_bindings

        key_bindings_instance = essential_modules["key_bindings_instance"]
        settings_manager_instance.override_key_bindings.assert_called_with(
            manager._handlers, key_bindings_instance, False
        )

    def test_toggle_enabled_with_minimal_script_interface(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test toggle_enabled method with script that has minimal interface."""

        self._setup_dependencies(test_context)
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        mock_script = test_context.Mock(spec=[])
        mock_script.add_key_grabs = test_context.Mock()
        mock_script.remove_key_grabs = test_context.Mock()
        mock_script.present_message = test_context.Mock()
        mock_script.key_bindings = test_context.Mock()
        mock_script.key_bindings.add = test_context.Mock()

        result = manager.toggle_enabled(mock_script)
        assert result is True
        assert manager.is_active() is True



# Unit tests for caret_navigator.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements
# pylint: disable=protected-access
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-locals

"""Unit tests for caret_navigator.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext


@pytest.mark.unit
class TestCaretNavigator:
    """Test CaretNavigator class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, Any]:
        """Set up mocks for caret_navigator dependencies."""

        essential_modules = test_context.setup_shared_dependencies()
        additional_modules = [
            "orca.input_event_manager",
            "orca.keybindings",
            "orca.cmdnames",
            "orca.guilabels",
            "orca.debug",
            "orca.ax_object",
            "orca.ax_text",
            "orca.script_manager",
        ]

        for module_name in additional_modules:
            if module_name not in essential_modules:
                essential_modules[module_name] = test_context.Mock()

        return essential_modules

    @pytest.mark.parametrize(
        "direction,event_provided,context_available,expected_result",
        [
            pytest.param("next", False, True, True, id="next_char_no_event_returns_true"),
            pytest.param("next", True, False, False, id="next_char_no_context_returns_false"),
            pytest.param("next", True, True, True, id="next_char_valid_navigation_succeeds"),
            pytest.param("previous", False, True, True, id="prev_char_no_event_returns_true"),
            pytest.param("previous", True, False, False, id="prev_char_no_context_returns_false"),
            pytest.param("previous", True, True, True, id="prev_char_valid_navigation_succeeds"),
        ],
    )
    def test_character_navigation(
        self,
        test_context: OrcaTestContext,
        direction: str,
        event_provided: bool,
        context_available: bool,
        expected_result: bool,
    ) -> None:
        """Test character navigation (next/previous) with various conditions."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        ax_object_mock = essential_modules["orca.ax_object"]
        ax_object_mock.AXObject.supports_text.side_effect = lambda obj: obj is not None
        ax_object_mock.AXObject.is_valid.side_effect = lambda obj: obj is not None
        ax_object_mock.AXObject.is_ancestor.side_effect = (
            lambda obj, root, same: obj is not None and root is not None
        )

        navigator = CaretNavigator()
        test_context.patch_object(navigator, "_get_root_object", return_value=None)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock() if event_provided else None

        if context_available:
            mock_obj = test_context.Mock()
            if direction == "next":
                mock_script.utilities.next_context.return_value = (mock_obj, 10)
            else:
                mock_script.utilities.previous_context.return_value = (mock_obj, 5)
        else:
            if direction == "next":
                mock_script.utilities.next_context.return_value = (None, 0)
            else:
                mock_script.utilities.previous_context.return_value = (None, 0)

        navigation_method = getattr(navigator, f"{direction}_character")
        result = navigation_method(mock_script, mock_event)
        assert result == expected_result

        if expected_result:
            assert navigator._last_input_event == mock_event
            mock_script.utilities.set_caret_position.assert_called_once()
            mock_script.interrupt_presentation.assert_called_once()
            mock_script.update_braille.assert_called_once()
            mock_script.say_character.assert_called_once()

    @pytest.mark.parametrize(
        "direction,context_result,word_contents,expected_result",
        [
            pytest.param("next", (None, 0), None, False, id="next_word_no_context"),
            pytest.param("next", ("obj", 20), [], False, id="next_word_no_contents"),
            pytest.param(
                "next", ("obj", 20), [("obj", 20, 25, "word")], True, id="next_word_success"
            ),
            pytest.param("previous", (None, 0), None, False, id="previous_word_no_context"),
            pytest.param("previous", ("obj", 15), [], False, id="previous_word_no_contents"),
            pytest.param(
                "previous", ("obj", 15), [("obj", 10, 15, "word")], True, id="previous_word_success"
            ),
        ],
    )
    def test_word_navigation(
        self,
        test_context: OrcaTestContext,
        direction: str,
        context_result: tuple,
        word_contents: list | None,
        expected_result: bool,
    ) -> None:
        """Test word navigation (next/previous) with various error conditions."""

        self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()

        if direction == "next":
            mock_script.utilities.next_context.return_value = context_result
        else:
            mock_script.utilities.previous_context.return_value = context_result

        mock_script.utilities.get_word_contents_at_offset.return_value = word_contents or []

        mock_script.utilities.set_caret_position = test_context.Mock()
        mock_script.interrupt_presentation = test_context.Mock()
        mock_script.update_braille = test_context.Mock()
        mock_script.say_word = test_context.Mock()

        navigation_method = getattr(navigator, f"{direction}_word")
        result = navigation_method(mock_script, mock_event)

        assert result == expected_result

        if expected_result:
            assert navigator._last_input_event == mock_event
            mock_script.utilities.set_caret_position.assert_called()
            mock_script.interrupt_presentation.assert_called_once()
            mock_script.update_braille.assert_called_once()
            mock_script.say_word.assert_called_once()
        else:
            mock_script.utilities.set_caret_position.assert_not_called()
            mock_script.interrupt_presentation.assert_not_called()

    @pytest.mark.parametrize(
        "test_method,expected_result",
        [
            pytest.param("suspend_commands", True, id="suspend_commands"),
            pytest.param("toggle_enabled", True, id="toggle_enabled"),
        ],
    )
    def test_navigator_control_methods(
        self,
        test_context: OrcaTestContext,
        test_method: str,
        expected_result: bool,
    ) -> None:
        """Test CaretNavigator control methods."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()

        if test_method == "suspend_commands":
            mock_script.key_bindings = test_context.Mock()
            mock_script.key_bindings.remove = test_context.Mock()
            mock_script.key_bindings.add = test_context.Mock()
            test_context.patch_object(navigator, "refresh_bindings_and_grabs")
            test_context.patch_object(navigator, "_is_active_script", return_value=True)
            navigator._suspended = False
            navigator.suspend_commands(mock_script, True, "test reason")
            assert navigator._suspended == expected_result

        elif test_method == "toggle_enabled":
            mock_script.key_bindings = test_context.Mock()
            mock_script.key_bindings.remove = test_context.Mock()
            mock_script.key_bindings.add = test_context.Mock()

            settings_manager_mock = essential_modules["orca.settings_manager"]
            manager_instance = test_context.Mock()
            settings_manager_mock.get_manager.return_value = manager_instance

            guilabels_mock = essential_modules["orca.guilabels"]
            guilabels_mock.CARET_NAVIGATION_ENABLED = "Caret navigation enabled"
            guilabels_mock.CARET_NAVIGATION_DISABLED = "Caret navigation disabled"

            test_context.patch_object(navigator, "refresh_bindings_and_grabs")

            result = navigator.toggle_enabled(mock_script, mock_event)
            assert result == expected_result

    @pytest.mark.parametrize(
        "test_method,param_data,expected_checks",
        [
            pytest.param(
                "initialization",
                {},
                ["_handlers", "_bindings", "_last_input_event", "_suspended"],
                id="initialization",
            ),
            pytest.param(
                "get_handlers",
                {"refresh": True, "expected_type": dict},
                ["isinstance_dict"],
                id="get_handlers_refresh",
            ),
            pytest.param(
                "get_handlers",
                {"refresh": False, "expected_type": dict},
                ["isinstance_dict"],
                id="get_handlers_cached",
            ),
            pytest.param(
                "get_bindings",
                {"is_desktop": True, "refresh": True},
                ["not_none"],
                id="get_bindings_desktop",
            ),
            pytest.param(
                "get_bindings",
                {"is_desktop": False, "refresh": True},
                ["not_none"],
                id="get_bindings_laptop",
            ),
        ],
    )
    def test_navigator_setup_and_getter_methods(
        self,
        test_context: OrcaTestContext,
        test_method: str,
        param_data: dict,
        expected_checks: list[str],
    ) -> None:
        """Test CaretNavigator setup and getter methods."""
        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()

        if test_method == "initialization":
            for attr in expected_checks:
                assert hasattr(navigator, attr)
            assert navigator._handlers is not None
            assert navigator._bindings is not None
            assert navigator._last_input_event is None
            assert navigator._suspended is False

        elif test_method == "get_handlers":
            handlers = navigator.get_handlers(refresh=param_data["refresh"])
            if "isinstance_dict" in expected_checks:
                assert isinstance(handlers, param_data["expected_type"])

        elif test_method == "get_bindings":
            keybindings_mock = essential_modules["orca.keybindings"]
            mock_bindings = test_context.Mock()
            keybindings_mock.KeyBindings = test_context.Mock(return_value=mock_bindings)

            result = navigator.get_bindings(
                refresh=param_data["refresh"], is_desktop=param_data["is_desktop"]
            )
            if "not_none" in expected_checks:
                assert result is not None

    @pytest.mark.parametrize(
        "navigation_type,in_say_all,current_line,next_prev_contents,expected_result",
        [
            pytest.param(
                "next_line", True, [("obj", 0, 10, "text")], [], True, id="next_line_in_say_all"
            ),
            pytest.param("next_line", False, [], [], False, id="next_line_no_current_line"),
            pytest.param(
                "next_line",
                False,
                [("obj", 0, 10, "text")],
                [],
                False,
                id="next_line_no_next_contents",
            ),
            pytest.param(
                "next_line",
                False,
                [("obj", 0, 10, "text")],
                [("obj2", 11, 21, "next")],
                True,
                id="next_line_success",
            ),
            pytest.param("previous_line", True, None, [], True, id="previous_line_in_say_all"),
            pytest.param("previous_line", False, None, [], False, id="previous_line_no_contents"),
            pytest.param(
                "previous_line",
                False,
                None,
                [("obj", 0, 10, "prev")],
                True,
                id="previous_line_success",
            ),
        ],
    )
    def test_line_navigation(
        self,
        test_context: OrcaTestContext,
        navigation_type: str,
        in_say_all: bool,
        current_line: list | None,
        next_prev_contents: list,
        expected_result: bool,
    ) -> None:
        """Test line navigation including say-all mode handling."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()

        focus_manager_mock = essential_modules["orca.focus_manager"]
        manager_instance = test_context.Mock()
        focus_manager_mock.get_manager.return_value = manager_instance
        manager_instance.in_say_all.return_value = in_say_all

        if in_say_all:
            settings_manager_mock = essential_modules["orca.settings_manager"]
            settings_instance = test_context.Mock()
            settings_manager_mock.get_manager.return_value = settings_instance
            settings_instance.get_setting.return_value = True

        if navigation_type == "next_line" and not in_say_all:
            mock_script.utilities.get_caret_context.return_value = ("obj", 5)
            mock_script.utilities.get_line_contents_at_offset.return_value = current_line
            mock_script.utilities.get_next_line_contents.return_value = next_prev_contents
        elif navigation_type == "previous_line" and not in_say_all:
            mock_script.utilities.get_caret_context.return_value = ("obj", 5)
            mock_script.utilities.get_previous_line_contents.return_value = next_prev_contents

        mock_script.utilities.set_caret_position = test_context.Mock()
        mock_script.interrupt_presentation = test_context.Mock()
        mock_script.speak_contents = test_context.Mock()
        mock_script.display_contents = test_context.Mock()

        navigation_method = getattr(navigator, f"{navigation_type}")
        result = navigation_method(mock_script, mock_event)

        assert result == expected_result

        if expected_result and not in_say_all:
            assert navigator._last_input_event == mock_event
            mock_script.utilities.set_caret_position.assert_called()
            mock_script.interrupt_presentation.assert_called_once()
            mock_script.speak_contents.assert_called_once()
            mock_script.display_contents.assert_called_once()
        elif in_say_all:
            assert navigator._last_input_event != mock_event

    @pytest.mark.parametrize(
        "navigation_type,line_contents,expected_result",
        [
            pytest.param("start_of_line", [], False, id="start_of_line_no_line"),
            pytest.param(
                "start_of_line", [("obj", 5, 15, "text")], True, id="start_of_line_success"
            ),
            pytest.param("end_of_line", [], False, id="end_of_line_no_line"),
            pytest.param(
                "end_of_line", [("obj", 5, 15, "text ")], True, id="end_of_line_with_space"
            ),
            pytest.param("end_of_line", [("obj", 5, 15, "text")], True, id="end_of_line_no_space"),
        ],
    )
    def test_line_boundary_navigation(
        self,
        test_context: OrcaTestContext,
        navigation_type: str,
        line_contents: list,
        expected_result: bool,
    ) -> None:
        """Test start/end of line navigation."""

        self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()

        mock_script.utilities.get_caret_context.return_value = ("obj", 10)
        mock_script.utilities.get_line_contents_at_offset.return_value = line_contents

        mock_script.utilities.set_caret_position = test_context.Mock()
        mock_script.interrupt_presentation = test_context.Mock()
        mock_script.say_character = test_context.Mock()
        mock_script.display_contents = test_context.Mock()

        navigation_method = getattr(navigator, f"{navigation_type}")
        result = navigation_method(mock_script, mock_event)

        assert result == expected_result

        if expected_result:
            assert navigator._last_input_event == mock_event
            mock_script.utilities.set_caret_position.assert_called()
            mock_script.interrupt_presentation.assert_called_once()
            mock_script.say_character.assert_called_once()
            mock_script.display_contents.assert_called_once()

    @pytest.mark.parametrize(
        "script_is_active,expected_result",
        [
            pytest.param(True, True, id="script_is_active"),
            pytest.param(False, False, id="script_is_not_active"),
        ],
    )
    def test_is_active_script(
        self,
        test_context: OrcaTestContext,
        script_is_active: bool,
        expected_result: bool,
    ) -> None:
        """Test _is_active_script method with active and non-active scripts."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_active_script = test_context.Mock() if script_is_active else test_context.Mock()

        script_manager_mock = essential_modules["orca.script_manager"]
        manager_instance = test_context.Mock()
        script_manager_mock.get_manager.return_value = manager_instance

        if script_is_active:
            manager_instance.get_active_script.return_value = mock_script
        else:
            manager_instance.get_active_script.return_value = mock_active_script

        debug_mock = essential_modules["orca.debug"]
        debug_mock.LEVEL_INFO = 800
        debug_mock.print_tokens = test_context.Mock()

        result = navigator._is_active_script(mock_script)
        assert result == expected_result

        if not script_is_active:
            debug_mock.print_tokens.assert_called_once()

    @pytest.mark.parametrize(
        "script_is_active,suspended,debug_level",
        [
            pytest.param(True, False, 0, id="active_script_not_suspended_debug_on"),
            pytest.param(True, True, 0, id="active_script_suspended_debug_on"),
            pytest.param(False, False, 0, id="non_active_script_debug_on"),
            pytest.param(True, False, 1000, id="active_script_debug_off"),
        ],
    )
    def test_add_bindings(
        self,
        test_context: OrcaTestContext,
        script_is_active: bool,
        suspended: bool,
        debug_level: int,
    ) -> None:
        """Test add_bindings method with various script states."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_script.key_bindings = test_context.Mock()
        mock_script.key_bindings.add = test_context.Mock()
        mock_script.key_bindings.get_bindings_with_grabs_for_debugging = test_context.Mock(
            return_value=[1, 2, 3]
        )

        navigator._suspended = suspended

        script_manager_mock = essential_modules["orca.script_manager"]
        manager_instance = test_context.Mock()
        script_manager_mock.get_manager.return_value = manager_instance

        if script_is_active:
            manager_instance.get_active_script.return_value = mock_script
        else:
            manager_instance.get_active_script.return_value = test_context.Mock()

        debug_mock = essential_modules["orca.debug"]
        debug_mock.LEVEL_INFO = 800
        debug_mock.debugLevel = debug_level
        debug_mock.print_tokens = test_context.Mock()

        mock_bindings = test_context.Mock()
        mock_bindings.key_bindings = [test_context.Mock(), test_context.Mock()]
        test_context.patch_object(navigator, "get_handlers", return_value={})
        test_context.patch_object(navigator, "get_bindings", return_value=mock_bindings)

        navigator.add_bindings(mock_script, "test reason")

        if script_is_active:
            assert debug_mock.print_tokens.call_count >= 1
            assert mock_script.key_bindings.add.call_count == len(mock_bindings.key_bindings)
            if debug_level <= 800:
                grabs_call_count = 2
                grabs_method = mock_script.key_bindings.get_bindings_with_grabs_for_debugging
                actual_grabs_call_count = grabs_method.call_count
                assert actual_grabs_call_count == grabs_call_count
        else:
            assert mock_script.key_bindings.add.call_count == 0

    @pytest.mark.parametrize(
        "script_is_active,debug_level",
        [
            pytest.param(True, 0, id="active_script_debug_on"),
            pytest.param(False, 0, id="non_active_script_debug_on"),
            pytest.param(True, 1000, id="active_script_debug_off"),
        ],
    )
    def test_remove_bindings(
        self,
        test_context: OrcaTestContext,
        script_is_active: bool,
        debug_level: int,
    ) -> None:
        """Test remove_bindings method with various script states."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_script.key_bindings = test_context.Mock()
        mock_script.key_bindings.remove = test_context.Mock()
        mock_script.key_bindings.get_bindings_with_grabs_for_debugging = test_context.Mock(
            return_value=[1, 2, 3]
        )

        mock_bindings = test_context.Mock()
        mock_bindings.key_bindings = [test_context.Mock(), test_context.Mock()]
        navigator._bindings = mock_bindings

        script_manager_mock = essential_modules["orca.script_manager"]
        manager_instance = test_context.Mock()
        script_manager_mock.get_manager.return_value = manager_instance

        if script_is_active:
            manager_instance.get_active_script.return_value = mock_script
        else:
            manager_instance.get_active_script.return_value = test_context.Mock()

        debug_mock = essential_modules["orca.debug"]
        debug_mock.LEVEL_INFO = 800
        debug_mock.debugLevel = debug_level
        debug_mock.print_tokens = test_context.Mock()

        navigator.remove_bindings(mock_script, "test reason")

        if script_is_active:
            assert debug_mock.print_tokens.call_count >= 1
            assert mock_script.key_bindings.remove.call_count == len(mock_bindings.key_bindings)
            if debug_level <= 800:
                grabs_call_count = 2
                grabs_method = mock_script.key_bindings.get_bindings_with_grabs_for_debugging
                actual_grabs_call_count = grabs_method.call_count
                assert actual_grabs_call_count == grabs_call_count
        else:
            assert mock_script.key_bindings.remove.call_count == 0

    def test_refresh_bindings_updates_script_bindings(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test refresh_bindings_and_grabs properly updates script key bindings."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_script.key_bindings = test_context.Mock()
        mock_script.key_bindings.remove = test_context.Mock()
        mock_script.key_bindings.add = test_context.Mock()
        mock_script.key_bindings.get_bindings_with_grabs_for_debugging = test_context.Mock(
            return_value=[1, 2, 3]
        )

        script_manager_mock = essential_modules["orca.script_manager"]
        manager_instance = test_context.Mock()
        script_manager_mock.get_manager.return_value = manager_instance
        manager_instance.get_active_script.return_value = mock_script

        debug_mock = essential_modules["orca.debug"]
        debug_mock.LEVEL_INFO = 800
        debug_mock.debugLevel = 0
        debug_mock.print_message = test_context.Mock()
        debug_mock.print_tokens = test_context.Mock()

        old_bindings = test_context.Mock()
        old_bindings.key_bindings = [test_context.Mock()]
        navigator._bindings = old_bindings

        new_bindings = test_context.Mock()
        new_bindings.key_bindings = [test_context.Mock(), test_context.Mock()]
        test_context.patch_object(navigator, "get_handlers", return_value={})
        test_context.patch_object(navigator, "get_bindings", return_value=new_bindings)

        navigator.refresh_bindings_and_grabs(mock_script, "test reason")

        assert mock_script.key_bindings.remove.call_count == len(old_bindings.key_bindings)
        assert mock_script.key_bindings.add.call_count == len(new_bindings.key_bindings)
        assert navigator._handlers == {}
        assert navigator._bindings == new_bindings
        debug_mock.print_message.assert_called_once()

    def test_get_is_enabled(self, test_context: OrcaTestContext) -> None:
        """Test get_is_enabled returns setting value."""

        essential_modules = self._setup_dependencies(test_context)
        settings_mock = test_context.Mock()
        settings_mock.get_setting.return_value = True
        essential_modules["orca.settings_manager"].get_manager.return_value = settings_mock
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.get_is_enabled()
        assert result is True
        settings_mock.get_setting.assert_called_with("caretNavigationEnabled")

    def test_set_is_enabled_no_change(self, test_context: OrcaTestContext) -> None:
        """Test set_is_enabled returns early if value unchanged."""

        essential_modules = self._setup_dependencies(test_context)
        settings_mock = test_context.Mock()
        settings_mock.get_setting.return_value = True
        essential_modules["orca.settings_manager"].get_manager.return_value = settings_mock
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.set_is_enabled(True)
        assert result is True
        settings_mock.set_setting.assert_not_called()

    def test_set_is_enabled_updates_setting(self, test_context: OrcaTestContext) -> None:
        """Test set_is_enabled updates setting and refreshes bindings."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        settings_mock = test_context.Mock()
        settings_mock.get_setting.return_value = False
        essential_modules["orca.settings_manager"].get_manager.return_value = settings_mock
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = mock_script
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_refresh = test_context.Mock()
        test_context.patch_object(navigator, "refresh_bindings_and_grabs", new=mock_refresh)

        result = navigator.set_is_enabled(True)
        assert result is True
        settings_mock.set_setting.assert_called_with("caretNavigationEnabled", True)
        assert navigator._last_input_event is None
        mock_refresh.assert_called_once()

    def test_set_is_enabled_no_active_script(self, test_context: OrcaTestContext) -> None:
        """Test set_is_enabled returns early if no active script."""

        essential_modules = self._setup_dependencies(test_context)
        settings_mock = test_context.Mock()
        settings_mock.get_setting.return_value = False
        essential_modules["orca.settings_manager"].get_manager.return_value = settings_mock
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = None
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_refresh = test_context.Mock()
        test_context.patch_object(navigator, "refresh_bindings_and_grabs", new=mock_refresh)

        result = navigator.set_is_enabled(True)
        assert result is True
        mock_refresh.assert_not_called()

    def test_get_triggers_focus_mode(self, test_context: OrcaTestContext) -> None:
        """Test get_triggers_focus_mode returns setting value."""

        essential_modules = self._setup_dependencies(test_context)
        settings_mock = test_context.Mock()
        settings_mock.get_setting.return_value = False
        essential_modules["orca.settings_manager"].get_manager.return_value = settings_mock
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.get_triggers_focus_mode()
        assert result is False
        settings_mock.get_setting.assert_called_with("caretNavTriggersFocusMode")

    def test_set_triggers_focus_mode(self, test_context: OrcaTestContext) -> None:
        """Test set_triggers_focus_mode updates setting."""

        essential_modules = self._setup_dependencies(test_context)
        settings_mock = test_context.Mock()
        settings_mock.get_setting.return_value = True
        essential_modules["orca.settings_manager"].get_manager.return_value = settings_mock
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.set_triggers_focus_mode(False)
        assert result is True
        settings_mock.set_setting.assert_called_with("caretNavTriggersFocusMode", False)

    def test_set_triggers_focus_mode_no_change(self, test_context: OrcaTestContext) -> None:
        """Test set_triggers_focus_mode returns early if unchanged."""

        essential_modules = self._setup_dependencies(test_context)
        settings_mock = test_context.Mock()
        settings_mock.get_setting.return_value = True
        essential_modules["orca.settings_manager"].get_manager.return_value = settings_mock
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.set_triggers_focus_mode(True)
        assert result is True
        settings_mock.set_setting.assert_not_called()

    def test_get_enabled_for_script(self, test_context: OrcaTestContext) -> None:
        """Test get_enabled_for_script returns script-specific state."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        navigator._enabled_for_script[mock_script] = True
        result = navigator.get_enabled_for_script(mock_script)
        assert result is True

    def test_get_enabled_for_script_default(self, test_context: OrcaTestContext) -> None:
        """Test get_enabled_for_script returns False by default."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.get_enabled_for_script(mock_script)
        assert result is False

    def test_set_enabled_for_script(self, test_context: OrcaTestContext) -> None:
        """Test set_enabled_for_script updates script-specific state."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        settings_mock = test_context.Mock()
        settings_mock.get_setting.return_value = False
        essential_modules["orca.settings_manager"].get_manager.return_value = settings_mock
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = mock_script
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_refresh = test_context.Mock()
        test_context.patch_object(navigator, "refresh_bindings_and_grabs", new=mock_refresh)
        test_context.patch_object(navigator, "_is_active_script", return_value=True)

        navigator.set_enabled_for_script(mock_script, True)
        assert navigator._enabled_for_script[mock_script] is True

    def test_set_enabled_for_script_inactive_script(self, test_context: OrcaTestContext) -> None:
        """Test set_enabled_for_script doesn't call set_is_enabled for inactive script."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        test_context.patch_object(navigator, "_is_active_script", return_value=False)
        mock_set_is_enabled = test_context.Mock()
        test_context.patch_object(navigator, "set_is_enabled", new=mock_set_is_enabled)

        navigator.set_enabled_for_script(mock_script, True)
        assert navigator._enabled_for_script[mock_script] is True
        mock_set_is_enabled.assert_not_called()

    def test_last_command_prevents_focus_mode_true(self, test_context: OrcaTestContext) -> None:
        """Test last_command_prevents_focus_mode returns True."""

        essential_modules = self._setup_dependencies(test_context)
        settings_mock = test_context.Mock()
        settings_mock.get_setting.return_value = False
        essential_modules["orca.settings_manager"].get_manager.return_value = settings_mock
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_event = test_context.Mock()
        navigator._last_input_event = mock_event
        test_context.patch_object(
            navigator, "last_input_event_was_navigation_command", return_value=True
        )
        result = navigator.last_command_prevents_focus_mode()
        assert result is True

    def test_last_command_prevents_focus_mode_false_no_event(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test last_command_prevents_focus_mode returns False if no event."""

        self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        navigator._last_input_event = None
        result = navigator.last_command_prevents_focus_mode()
        assert result is False

    def test_last_command_prevents_focus_mode_false_setting_true(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test last_command_prevents_focus_mode returns False if setting True."""

        essential_modules = self._setup_dependencies(test_context)
        settings_mock = test_context.Mock()
        settings_mock.get_setting.return_value = True
        essential_modules["orca.settings_manager"].get_manager.return_value = settings_mock
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_event = test_context.Mock()
        navigator._last_input_event = mock_event
        test_context.patch_object(
            navigator, "last_input_event_was_navigation_command", return_value=True
        )
        result = navigator.last_command_prevents_focus_mode()
        assert result is False
